# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ...model import implements, Model

from .. import ICodec, IIoManager, IStorage, IStorageFactory


@implements(IIoManager)
class IoManager(Model):
    """ The IoManager class is the default implementation of the
    :class:`~dip.io.IIoManager` interface.
    """

    def readable_locations_from_string(self, format, location):
        """ Get the list of readable storage location instances for which a
        location specified as a string is valid.

        :param format:
            is the identifier of the format.
        :param location:
            is the location as a string.
        :return:
            the list of storage location instances.
        """

        locations = []
        for storage in self.readable_storage(format):
            storage_location = storage.readable_location(location)
            if storage_location is not None:
                locations.append(storage_location)

        return locations

    def readable_storage(self, format):
        """ Get the list of the storage instances that can be read from using a
        particular format.

        :param format:
            is the identifier of the format.
        :return:
            the list of storage instances.
        """

        return self._storage_for_format(format, reading=True)

    def writeable_locations_from_string(self, format, location):
        """ Get the list of writeable storage location instances for which a
        location specified as a string is valid.

        :param format:
            is the identifier of the format.
        :param location:
            is the location as a string.
        :return:
            the list of storage location instances.
        """

        locations = []
        for storage in self.writeable_storage(format):
            storage_location = storage.writeable_location(location)
            if storage_location is not None:
                locations.append(storage_location)

        return locations

    def writeable_storage(self, format):
        """ Get the list of the storage instances that can be written to using
        a particular format.

        :param format:
            is the identifier of the format.
        :return:
            the list of storage instances.
        """

        return self._storage_for_format(format, reading=False)

    @IIoManager.ui.default
    def ui(self):
        """ Return the default implementation of the storage-independent user
        interfaces.
        """

        from .io_manager_ui import IoManagerUi

        return IoManagerUi(io_manager=self)

    def _storage_for_format(self, format, reading):
        """ Return a list of storage instances that can handle a format. """

        storage_list = []

        # If we have at exactly one usable codec then we can use any streaming
        # storage (subject to any storage policy).
        streaming_codec = None

        for codec in self.codecs:
            if self._invalid_codec(codec, format, reading):
                continue

            if streaming_codec is not None:
                # Note that this restriction doesn't apply to structured
                # storage.
                raise ValueError(
                        "multiple codecs have been defined for format "
                        "'{0}'".format(format))

            streaming_codec = codec

        for storage_factory in self.storage_factories:
            istoragefactory = IStorageFactory(storage_factory)

            usable = istoragefactory.readable if reading else istoragefactory.writeable
            if not usable:
                continue

            if isinstance(storage_factory, ICodec):
                if self._invalid_codec(storage_factory, format, reading):
                    continue

                codec = storage_factory
            elif streaming_codec is None:
                continue
            else:
                codec = streaming_codec

            # Create the storage instance bound to the codec.
            storage = storage_factory(codec)

            # Apply any storage policies.
            for storage_policy in self.storage_policies:
                if not storage_policy(format, storage):
                    break
            else:
                storage_list.append(storage)

        return storage_list

    @staticmethod
    def _invalid_codec(codec, format, reading):
        """ Check the validity of a codec. """

        icodec = ICodec(codec)

        if icodec.format != format:
            return True

        iface = icodec.decoder_interface if reading else icodec.encoder_interface

        return iface is None
