# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ...model import Any, Bool, Dict, Instance, List, observe
from ...ui import IViewStack, Wizard, IWizard, WizardController, IWizardPage

from .. import IIoManagerUi, IStorage, IStorageBrowser, IStorageLocation


class LocationWizard(Wizard):
    """ LocationWizard is an internal class that implements the wizard needed
    when the user has to specify both the storage and the location within the
    chosen storage.
    """

    # Set if the wizard is asking for readable storage.
    reading = Bool(True)

    # The hints.
    hints = Any()

    # The storage location.
    location = Instance(IStorageLocation)

    # The current storage.
    storage = Instance(IStorage)

    # The list of storage instances.
    storage_list = List(IStorage)

    # The i/o manager user interface using the wizard.
    ui = Instance(IIoManagerUi)

    def __init__(self):
        """ Initialise the wizard. """

        iiomanagerui = IIoManagerUi(self.ui)

        if self.reading:
            storage_page = iiomanagerui.readable_storage_wizard_page(
                    'storage', storage_list=self.storage_list)

            location_page = iiomanagerui.readable_location_wizard_page(
                    id='dip.io.location.location_page')
        else:
            storage_page = iiomanagerui.writeable_storage_wizard_page(
                    'storage', storage_list=self.storage_list)

            location_page = iiomanagerui.writeable_location_wizard_page(
                    id='dip.io.location.location_page')

        super(LocationWizard, self).__init__(storage_page, location_page)

    def execute(self, parent):
        """ Create and execute the wizard.  Return True if the user accepted.
        """

        # Create browsers for the storage and determine the current storage if
        # there is an initial location.
        browsers = {}
        browser_list = []

        for storage in self.storage_list:
            if self.location is not None and self.location.storage is storage:
                self.storage = storage

            browser = storage.ui.write_browser(default_location=self.location,
                    hints=self.hints)
            browsers[storage] = browser
            browser_list.append(browser)

        controller = _WizardController(browsers=browsers)

        wizard = self(self, controller=controller, parent=parent)

        iwizard = IWizard(wizard)

        IViewStack(IWizardPage(iwizard.views[1]).view).views = browser_list

        return iwizard.execute()


class _WizardController(WizardController):
    """ An internal class that updates the contents of various pages of the
    location wizard that are dependent on the contents of other pages.
    """

    # Automatically update the model.
    auto_update_model = True

    # The browsers keyed by storage.
    browsers = Dict()

    @observe('current_page')
    def _page_changed(self, change):
        """ Invoked when the wizard's current page changes. """

        if change.new is not None:
            ipage = IWizardPage(change.new)

            if ipage.id == 'dip.io.location.location_page':
                # Update the storage browser now that we know what storage has
                # been selected.
                IViewStack(ipage.view).current_view = self._current_browser()

    def validate_page(self, page):
        """ Reimplemented to validate the location page differently. """

        ipage = IWizardPage(page)

        if ipage.id != 'dip.io.location.location_page':
            return super(_WizardController, self).validate_page(page)

        istoragebrowser = IStorageBrowser(self._current_browser())

        # It doesn't matter if it is invalid.
        self.model.location = istoragebrowser.location

        return istoragebrowser.invalid_reason

    @browsers.setter
    def browsers(self, value):
        """ Invoked to set the dictionary of browsers. """

        # Save for the standard getter.
        self._browsers = value

        # Re-validate when any location changes.
        for browser in value.values():
            observe('location', IStorageBrowser(browser), self.validate)

    def _current_browser(self):
        """ Return the current browser. """

        return self.browsers[self.model.storage]
