# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from .exceptions import ValidationError, ValidationTypeError
from .value_type_factory import ValueTypeFactory


class Enum(ValueTypeFactory):
    """ The Enum class encapsulates a fixed set of string values. """

    def __init__(self, *members, **metadata):
        """ Initialise the object.

        :param \*members:
            the list of members.
        :param default:
            the default attribute value.  If omitted then the first member is
            used.
        :param allow_none:
            ``True`` if ``None`` is a valid value.  The default is ``False``.
        :param getter:
            is the optional attribute getter.
        :param setter:
            is the optional attribute setter.
        :param \*\*metadata:
            is additional meta-data stored with the type.
        """

        default = metadata.pop('default', None)
        allow_none = metadata.pop('allow_none', False)
        getter = metadata.pop('getter', None)
        setter = metadata.pop('setter', None)

        if len(members) == 0:
            raise ValidationError(type(self), message="no members defined")

        for member in members:
            if not isinstance(member, str):
                raise ValidationTypeError(type(self), str, member)

        self.members = members

        if default is None and not allow_none:
            default = members[0]

        super(Enum, self).__init__(default, allow_none, getter, setter,
                metadata)

    def copy(self):
        """ Create a copy of this instance.

        :return:
            the copied instance.
        """

        return self.__class__(*self.members, default=self.default_value,
                allow_none=self.allow_none, getter=self.getter_func,
                setter=self.setter_func, **self.metadata)

    def validate(self, value):
        """ Validate an object according to the constraints.  An exception is
        raised if the object is invalid.

        :param value:
            the object to validate.
        :return:
            the validated object.
        """

        if value is None and self.allow_none:
            return value

        if value not in self.members:
            member_list = ', '.join(
                    ['\'' + member + '\'' for member in self.members])

            raise ValidationError(type(self),
                    message="expected one of {0}, not '{1}'".format(
                            member_list, value))

        return value
