# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from .any import Any
from .attribute_change import AttributeChange
from .change_trigger import get_change_trigger
from .exceptions import ValidationTypeError
from .mutable_type_factory import MutableTypeFactory


# A unique object that is used to determine whether an optional argument was
# given or not.
DefaultArgValue = object()


class List(MutableTypeFactory):
    """ The List class encapsulates a Python list with elements of a
    particular type.
    """

    def __init__(self, element_type=None, default=DefaultArgValue, allow_none=False, getter=None, setter=None, **metadata):
        """ Initialise the object.

        :param element_type:
            the type of each element of the list.  If it is omitted then
            elements of any type are allowed.  If it is a string then it is
            the "full" name of the type.  The string form allows types to be
            specified while avoiding circular imports.
        :param default:
            is the default value of the list.  If is it omitted then an empty
            list is the default.
        :param allow_none:
            ``True`` if ``None`` is a valid value.  The default is ``False``.
        :param getter:
            is the optional attribute getter.
        :param setter:
            is the optional attribute setter.
        :param \*\*metadata:
            is additional meta-data stored with the type.
        """

        self.element_type = element_type

        if default is DefaultArgValue:
            default = []

        super(List, self).__init__(default, allow_none, getter, setter,
                metadata)

    def copy(self):
        """ Create a copy of this instance.

        :return:
            the copied instance.
        """

        return self.__class__(self.element_type, self.default_value,
                self.allow_none, self.getter_func, self.setter_func,
                **self.metadata)

    def get_default_value(self):
        """ Reimplemented to return a copy of the default because lists are
        mutable.

        :return:
            a copy of the default value.
        """

        return ListImpl(self, self.default_value)

    def validate(self, value):
        """ Validate a list according to the constraints.  An exception is
        raised if the list is invalid.

        :param value:
            the list to validate.
        :return:
            the validated list.
        """

        if value is None and self.allow_none:
            return value

        self.element_type = self.resolve_type(self.element_type)

        # Allow any iterator.
        if not hasattr(value, '__iter__'):
            raise ValidationTypeError(type(self), list, value)

        return ListImpl(self, self.validated_list(value))

    def validated_list(self, value):
        """ Return a list containing validated elements.

        :param value:
            the list of (possibly) invalid elements.
        :return:
            a corresponding list of validated elements.
        """

        # There is no need to validate an untyped list.
        if self.element_type is None or isinstance(self.element_type, Any):
            return value

        # We only need to check the type of a validated list.
        if isinstance(value, ListImpl):
            self.validate_collection_type(self.element_type,
                    value._dip_element_type)

            return value

        return [self.validate_collection_value(self.element_type, v)
                for v in value]


class ListImpl(list):
    """ The ListImpl class is a list that validates new elements against a
    type.
    """

    def __init__(self, list_type, validated_list):
        """ Initialise the object. """

        self._dip_list_type = list_type
        self._dip_element_type = list_type.element_type
        self._dip_model = None

        super(ListImpl, self).__init__(validated_list)

    def _dip_changed(self, trigger, new, old):
        """ Notify any change. """

        if new != old:
            change = AttributeChange(model=self._dip_model,
                    name=self._dip_list_type.name, new=new, old=old)
            trigger.changed.emit(change)

    def _dip_change_trigger(self):
        """ Return the change trigger if there is one. """

        return get_change_trigger(self._dip_model, self._dip_list_type.name)

    def append(self, element):
        """ Reimplemented to validate the element. """

        trigger = self._dip_change_trigger()

        element = self._dip_list_type.validate_collection_value(
                self._dip_element_type, element)

        super(ListImpl, self).append(element)

        if trigger is not None:
            self._dip_changed(trigger, [element], [])

    def extend(self, extension):
        """ Reimplemented to validate the list extension. """

        trigger = self._dip_change_trigger()

        extension = self._dip_list_type.validated_list(extension)

        super(ListImpl, self).extend(extension)

        if trigger is not None:
            self._dip_changed(trigger, extension, [])

    def insert(self, n, element):
        """ Reimplemented to validate the element. """

        trigger = self._dip_change_trigger()

        element = self._dip_list_type.validate_collection_value(
                self._dip_element_type, element)

        super(ListImpl, self).insert(n, element)

        if trigger is not None:
            self._dip_changed(trigger, [element], [])

    def pop(self, n=None):
        """ Reimplemented to emit the change notification. """

        trigger = self._dip_change_trigger()

        if n is None:
            element = super(ListImpl, self).pop()
        else:
            element = super(ListImpl, self).pop(n)

        if trigger is not None:
            self._dip_changed(trigger, [], [element])

        return element

    def remove(self, element):
        """ Reimplemented to validate the element. """

        trigger = self._dip_change_trigger()

        element = self._dip_list_type.validate_collection_value(
                self._dip_element_type, element)

        super(ListImpl, self).remove(element)

        if trigger is not None:
            self._dip_changed(trigger, [], [element])

    def reverse(self):
        """ Reimplemented to emit the change notification. """

        trigger = self._dip_change_trigger()

        if trigger is not None:
            old = list(self)

        super(ListImpl, self).reverse()

        if trigger is not None:
            self._dip_changed(trigger, self, old)

    def sort(self, key=None, reverse=0):
        """ Reimplemented to emit the change notification. """

        trigger = self._dip_change_trigger()

        if trigger is not None:
            old = list(self)

        super(ListImpl, self).sort(key=key, reverse=reverse)

        if trigger is not None:
            self._dip_changed(trigger, self, old)

    def __iadd__(self, extension):
        """ Reimplemented to validate the list extension. """

        trigger = self._dip_change_trigger()

        if trigger is not None:
            old = list(self)

        extension = self._dip_list_type.validated_list(extension)

        ret = super(ListImpl, self).__iadd__(extension)

        if trigger is not None:
            self._dip_changed(trigger, ret, old)

        return ret

    def __setitem__(self, key, value):
        """ Reimplemented to validate the value. """

        trigger = self._dip_change_trigger()

        if isinstance(key, slice):
            value = self._dip_list_type.validated_list(value)

            if trigger is not None:
                old = self.__getitem__(key)
                new = value
        else:
            value = self._dip_list_type.validate_collection_value(
                    self._dip_element_type, value)

            if trigger is not None:
                old = [self.__getitem__(key)]
                new = [value]

        super(ListImpl, self).__setitem__(key, value)

        if trigger is not None:
            self._dip_changed(trigger, new, old)

    def __delitem__(self, n):
        """ Reimplemented to emit the change notification. """

        trigger = self._dip_change_trigger()

        if trigger is not None:
            if isinstance(n, slice):
                old = self.__getitem__(n)
            else:
                old = [self.__getitem__(n)]

        super(ListImpl, self).__delitem__(n)

        if trigger is not None:
            self._dip_changed(trigger, [], old)

    def __setslice__(self, i, j, value):
        """ Reimplemented to validate the value. """

        trigger = self._dip_change_trigger()

        value = self._dip_list_type.validated_list(value)

        if trigger is not None:
            old = self[i:j]

        super(ListImpl, self).__setslice__(i, j, value)

        if trigger is not None:
            self._dip_changed(trigger, value, old)

    def __delslice__(self, i, j):
        """ Reimplemented to emit the change notification. """

        trigger = self._dip_change_trigger()

        if trigger is not None:
            old = self[i:j]

        super(ListImpl, self).__delslice__(i, j)

        if trigger is not None:
            self._dip_changed(trigger, [], old)
