# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Model, Tuple


class BaseToolkit(Model):
    """ The BaseToolkit class is a base class for :term:`toolkits<toolkit>`
    that implements the default mapping of :mod:`dip.model` types to the
    :term:`editors<editor>` used to handle them.
    """

    # A map of declarative factories to attribute types.
    attribute_type_to_declarative_factory_map = Tuple()

    # A map of editor factories to attribute types.
    attribute_type_to_editor_factory_map = Tuple()

    # A map of object factories to interfaces.
    interface_to_object_factory_map = Tuple()

    def declarative_factory_for_attribute_type(self, attribute_type):
        """ Get an :class:`~dip.ui.EditorFactory` sub-class that will create a
        factory that will create an editor to handle attributes of a particular
        type.

        :param attribute_type:
            is the type of the attribute.
        :return:
            the :class:`~dip.ui.EditorFactory` sub-class.
        """

        for at, df in self.attribute_type_to_declarative_factory_map:
            if isinstance(attribute_type, at):
                return df

        raise TypeError(
                "there is no editor factory that creates instances of "
                "{0}".format(type(attribute_type).__name__))

    @attribute_type_to_declarative_factory_map.default
    def attribute_type_to_declarative_factory_map(self):
        """ Invoked to get the default map between attribute types and
        declarative factories.
        """

        from .. import model, ui

        return (
            (model.Bool,        ui.CheckBox),
            (model.Enum,        ui.ComboBox),
            (model.Float,       ui.FloatSpinBox),
            (model.Int,         ui.SpinBox),
            (model.List,        ui.ListEditor),
            (model.Str,         ui.LineEditor),
            (model.Trigger,     ui.PushButton)
        )

    def editor_factory_for_attribute_type(self, attribute_type):
        """ Get a callable that will create a editor to handle attributes of a
        particular type.

        :param attribute_type:
            is the type of the attribute.
        :return:
            the callable.
        """

        for at, ef in self.attribute_type_to_editor_factory_map:
            if isinstance(attribute_type, at):
                return ef

        raise TypeError(
                "there is no editor factory that creates instances of "
                "{0}".format(type(attribute_type).__name__))

    @attribute_type_to_editor_factory_map.default
    def attribute_type_to_editor_factory_map(self):
        """ Invoked to get the default map between attribute types and editor
        factories.
        """

        from .. import model, pui

        return (
            (model.Bool,        pui.CheckBox),
            (model.Enum,        pui.ComboBox),
            (model.Float,       pui.FloatSpinBox),
            (model.Int,         pui.SpinBox),
            (model.Str,         pui.LineEditor)
        )

    def object_factory_for_interface(self, interface):
        """ Get the factory that will create an object that is adapted to a
        particular interface.

        :param interface:
            is the interface.
        :return:
            a callable that will accept a single argument that is an optional
            parent view and will return an object that is adapted to a
            sub-class of the :class:`~dip.ui.IObject` interface.
        """

        for iface, of in self.interface_to_object_factory_map:
            if issubclass(interface, iface):
                return of

        raise TypeError(
                "there is no object factory that creates implementations of "
                "{0}".format(interface.__name__))

    @interface_to_object_factory_map.default
    def interface_to_object_factory_map(self):
        """ Invoked to get the default map between interfaces and object
        factories.
        """

        from .. import pui, ui

        return (
            (ui.IAction,                    pui.Action),
            (ui.ICheckBox,                  pui.CheckBox),
            (ui.IComboBox,                  pui.ComboBox),
            (ui.IDialog,                    pui.Dialog),
            (ui.IDock,                      pui.Dock),
            (ui.IFilesystemLocationEditor,  pui.FilesystemLocationEditor),
            (ui.IFloatSpinBox,              pui.FloatSpinBox),
            (ui.IForm,                      pui.Form),
            (ui.IGroupBox,                  pui.GroupBox),
            (ui.IHBox,                      pui.HBox),
            (ui.ILabel,                     pui.Label),
            (ui.ILineEditor,                pui.LineEditor),
            (ui.IListEditor,                pui.ListEditor),
            (ui.IMainWindow,                pui.MainWindow),
            (ui.IMenu,                      pui.Menu),
            (ui.IMenuBar,                   pui.MenuBar),
            (ui.IMessageArea,               pui.MessageArea),
            (ui.IOptionList,                pui.OptionList),
            (ui.IPushButton,                pui.PushButton),
            (ui.IRadioButtons,              pui.RadioButtons),
            (ui.ISpinBox,                   pui.SpinBox),
            (ui.ISplitter,                  pui.Splitter),
            (ui.IStorageLocationEditor,     pui.StorageLocationEditor),
            (ui.ITabBar,                    pui.TabBar),
            (ui.ITableEditor,               pui.TableEditor),
            (ui.ITextEditor,                pui.TextEditor),
            (ui.IToolButton,                pui.ToolButton),
            (ui.IVBox,                      pui.VBox),
            (ui.IViewStack,                 pui.ViewStack),
            (ui.IWizard,                    pui.Wizard),
            (ui.IWizardPage,                pui.WizardPage)
        )
