# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


import sys

from PyQt4.QtCore import QEvent, QObject, QPoint, Qt, QTimer
from PyQt4.QtGui import (QAbstractButton, QAbstractItemView, QApplication,
        QLayout, QTableWidget, QTableWidgetItem)
from PyQt4.QtTest import QTest

from ....automate import (AutomationError, IAutomatedEditor,
        IAutomatedTableEditor)
from ....model import adapt, Adapter, Instance, Model, notify_observers
from ....ui import ITableEditor

from ..editor_delegate import EditorDelegate
from ..editor_layout_adapter import EditorLayoutAdapter
from ..visual_items import find_visual_item


# Qt should handle this.
_EDIT_KEY = Qt.Key_Enter if sys.platform == 'darwin' else Qt.Key_F2


class TableEditorComponents(Model):
    """ The TableEditorComponents class is an internal mixin for adapters that
    access the embedded components of a table editor.
    """

    # The table widget.
    table_widget = Instance(QTableWidget)

    @table_widget.default
    def table_widget(self):
        """ Invoked to return the table widget. """

        return find_visual_item(self.adaptee, QTableWidget)

    # The add button.
    add_button = Instance(QAbstractButton)

    @add_button.default
    def add_button(self):
        """ Invoked to return the add button. """

        return find_visual_item(self.adaptee, QAbstractButton, name='add')

    # The remove button.
    remove_button = Instance(QAbstractButton)

    @remove_button.default
    def remove_button(self):
        """ Invoked to return the remove button. """

        return find_visual_item(self.adaptee, QAbstractButton, name='remove')

    @classmethod
    def usable_editor(cls, item):
        """ Return True if an item can be used as an editor. """

        # The table widget must be an immediate child otherwise the criteria
        # would be far too wide.
        if find_visual_item(item, QTableWidget, recurse=False) is None:
            return False

        if find_visual_item(item, QAbstractButton, name='add') is None:
            return False

        if find_visual_item(item, QAbstractButton, name='remove') is None:
            return False

        return True


@adapt(QLayout, to=ITableEditor)
class QLayoutITableEditorAdapter(EditorLayoutAdapter, TableEditorComponents):
    """ An adapter to implement ITableEditor for a layout that embeds a
    :class:`~PyQt4.QtGui.QTableWidget`, a :class:`~PyQt4.QtGui.QAbstractButton`
    named 'add' and a :class:`~PyQt4.QtGui.QAbstractButton` named 'remove'.
    """

    # The table widget's event filter.
    _event_filter = Instance(QObject)

    @classmethod
    def isadaptable(cls, adaptee):
        """ Determine if this adapter can adapt a particular layout.

        :param adaptee:
            is the layout.
        :return:
            ``True`` if the layout can be adapted.
        """

        return cls.usable_editor(adaptee)

    def configure(self, properties):
        """ Configure the editor. """

        table_widget = self.table_widget
        table_widget.setDragDropMode(QAbstractItemView.InternalMove)
        table_widget.installEventFilter(self._event_filter)
        table_widget.verticalHeader().hide()

        self.add_button.clicked[()].connect(self._on_add_clicked)

        remove_button = self.remove_button
        remove_button.setEnabled(False)
        remove_button.clicked[()].connect(self._on_remove_clicked)

        # Apply any properties to the table widget.
        try:
            table_widget.pyqtConfigure(**properties)
        except AttributeError:
            pass

    @ITableEditor.columns.setter
    def columns(self, value):
        """ Invoked to set the columns. """

        # Save for the standard getter.
        self._columns = value

        table_widget = self.table_widget

        # Set the number of columns.
        table_widget.setColumnCount(len(self.columns))

        # If there are any labels then use them for the column headings.
        labels = []
        has_labels = False

        for col_nr, col in enumerate(value):
            table_widget.setItemDelegateForColumn(col_nr,
                    EditorDelegate(col.editor_factory, col.column_type,
                            self.adaptee))

            labels.append(col.label)
            if col.label != '':
                has_labels = True

        if has_labels:
            table_widget.setHorizontalHeaderLabels(labels)
            table_widget.horizontalHeader().setStretchLastSection(True)
        else:
            table_widget.horizontalHeader().hide()

    @ITableEditor.read_only.getter
    def read_only(self):
        """ Invoked to get the read-only state. """

        return self.table_widget.editTriggers() == QAbstractItemView.NoEditTriggers

    @read_only.setter
    def read_only(self, value):
        """ Invoked to set the read-only state. """

        table_widget = self.table_widget

        table_widget.setEditTriggers(
                QAbstractItemView.NoEditTriggers if value else
                        QAbstractItemView.DoubleClicked |
                        QAbstractItemView.EditKeyPressed |
                        QAbstractItemView.AnyKeyPressed)

        table_widget.setDragDropMode(
                QAbstractItemView.NoDragDrop if value else
                        QAbstractItemView.InternalMove)

        self.remove_button.parent().setVisible(not value)

    @ITableEditor.value.getter
    def value(self):
        """ Invoked to get the editor's value. """

        table_widget = self.table_widget

        value = []

        for row_nr in range(table_widget.rowCount()):
            row = self.row_factory()

            for col_nr, col in enumerate(self.columns):
                itm = table_widget.item(row_nr, col_nr)

                # There may not be an item if it has been internally moved.
                if itm is None:
                    self._new_cell(row_nr, col_nr, col, row[col_nr])
                else:
                    cell = itm.data(Qt.EditRole).toPyObject()

                    # The data may be ``None`` if there is an open delegate.
                    if cell is not None:
                        row[col_nr] = cell

            value.append(row)

        return value

    @value.setter
    def value(self, value):
        """ Invoked to set the editor's value. """

        table_widget = self.table_widget

        # Save for the notifier.
        self.qt4_old_value = list(value)

        old_nr_rows = table_widget.rowCount()
        new_nr_rows = len(value)

        cols = range(table_widget.columnCount())

        # Update any rows that will be kept.
        keep_rows = min(old_nr_rows, new_nr_rows)

        for row_nr in range(keep_rows):
            row = value[row_nr]

            for col_nr in cols:
                itm = table_widget.item(row_nr, col_nr)
                itm.setData(Qt.EditRole, row[col_nr])
                itm.setData(Qt.UserRole, '')

        row_nr = keep_rows

        if old_nr_rows < new_nr_rows:
            table_widget.setRowCount(new_nr_rows)

            # Add the rows for the new elements.
            while row_nr < new_nr_rows:
                self._new_row(value[row_nr], row_nr)
                row_nr += 1

        elif old_nr_rows > new_nr_rows:
            # Remove the rows for the old elements.
            while row_nr < old_nr_rows:
                for col_nr in cols:
                    table_widget.takeItem(row_nr, col_nr)

                row_nr += 1

            table_widget.setRowCount(new_nr_rows)

        self.remove_button.setEnabled(new_nr_rows != 0)

    @ITableEditor.invalid_reason.getter
    def invalid_reason(self):
        """ Invoked to get the invalid reason. """

        table_widget = self.table_widget

        # Find an invalid item giving preference to the current cell.
        itm = table_widget.currentItem()

        reason = '' if itm is None else itm.data(Qt.UserRole).toPyObject()
        if reason == '':
            for row_nr in range(table_widget.rowCount()):
                for col_nr in range(table_widget.columnCount()):
                    itm = table_widget.item(row_nr, col_nr)
                    if itm is None:
                        continue

                    reason = itm.data(Qt.UserRole).toPyObject()
                    if reason != '':
                        break

        if reason != '':
            row_nr = table_widget.row(itm)

            if len(self.columns) == 1:
                reason = "in row {0}, {2}".format(row_nr + 1, reason)
            else:
                col_nr = table_widget.column(itm)

                col_itm = table_widget.horizontalHeaderItem(col_nr)
                if col_itm is None:
                    col_name = "column {0}".format(col_nr)
                else:
                    col_name = col_itm.text()

                reason = "in row {0} of '{1}', {2}".format(row_nr + 1,
                        col_name, reason)

        return reason

    def item_invalid_reason(self, invalid_reason):
        """ Update the invalid reason for the current item in the collection.

        :param index:
            is the invalid reason.  It will be an empty string if the item is
            valid.
        """

        self.table_widget.currentItem().setData(Qt.UserRole, invalid_reason)

    @_event_filter.default
    def _event_filter(self):
        """ Invoked to return the default event filter. """

        return EventFilter(self)

    def _on_add_clicked(self):
        """ Invoked to add a new row to the collection. """

        table_widget = self.table_widget

        # Add a new row.
        new_row_nr = table_widget.rowCount()
        table_widget.setRowCount(new_row_nr + 1)
        itm = self._new_row(self.row_factory(), new_row_nr)

        self.remove_button.setEnabled(True)

        # Start editing it.
        table_widget.setCurrentItem(itm)
        table_widget.editItem(itm)

    def _on_remove_clicked(self):
        """ Invoked to remove the current row from the collection. """

        table_widget = self.table_widget

        # Do this now because rowCount() doesn't get updated quickly enough
        # afterwards.
        last_row = (table_widget.rowCount() == 1)

        self._event_filter.setRemovingRows(True)
        table_widget.removeRow(table_widget.currentRow())
        self._event_filter.setRemovingRows(False)

        if last_row:
            self.remove_button.setEnabled(False)

        self.qt4_notify(self.value)

    def _new_row(self, row, row_nr):
        """ Add a new row at a particular row number. """

        first_itm = None

        for col_nr, col in enumerate(self.columns):
            itm = self._new_cell(row_nr, col_nr, col, row[col_nr])

            if first_itm is None:
                first_itm = itm

        return first_itm

    def _new_cell(self, row_nr, col_nr, col, cell):
        """ Create a new item for a cell. """

        itm = QTableWidgetItem()

        if not self.read_only and col.editable:
            itm.setFlags(itm.flags() | Qt.ItemIsEditable)

        itm.setData(Qt.EditRole, cell)
        itm.setData(Qt.UserRole, '')

        self.table_widget.setItem(row_nr, col_nr, itm)

        return itm

    def qt4_order_changed(self):
        """ Invoked when the order of items changes. """

        self.qt4_notify(self.value)


class EventFilter(QObject):
    """ EventFilter is an internal class that filters a table widget's events
    in order to detect internal moves.
    """

    def __init__(self, widget_adapter):
        """ Initialise the object. """

        super(EventFilter, self).__init__(widget_adapter.adaptee)

        self._removing_rows = False
        self._widget_adapter = widget_adapter

    def eventFilter(self, obj, event):
        """ Reimplemented to detect internal moves. """

        # We can't simply look for drop events (they don't actually occur).
        if not self._removing_rows and event.type() == QEvent.ChildRemoved:
            # Notify after the event has been fully handled.
            QTimer.singleShot(0, self._widget_adapter.qt4_order_changed)

        return super(EventFilter, self).eventFilter(obj, event)

    def setRemovingRows(self, removing):
        """ Remember if we are removing rows programmatically. """

        self._removing_rows = removing


@adapt(QLayout, to=IAutomatedTableEditor)
class QLayoutIAutomatedTableEditorAdapter(Adapter, TableEditorComponents):
    """ An adapter to implement IAutomatedTableEditor for a layout that embeds
    a :class:`~PyQt4.QtGui.QTableWidget`, a
    :class:`~PyQt4.QtGui.QAbstractButton` named 'add' and a
    :class:`~PyQt4.QtGui.QAbstractButton` named 'remove'.
    """

    @classmethod
    def isadaptable(cls, adaptee):
        """ Determine if this adapter can adapt a particular layout.

        :param adaptee:
            is the layout.
        :return:
            ``True`` if the layout can be adapted.
        """

        return cls.usable_editor(adaptee)

    def simulate_append(self, id, delay, values):
        """ Simulate the user appending a row to the table.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param values:
            is the sequence of values to append.
        """

        table_widget = self.table_widget

        nr_values = len(values)

        if nr_values == 0:
            raise AutomationError(id, 'append',
                    "at least one value must be given")

        nr_columns = table_widget.columnCount()

        if nr_values > nr_columns:
            raise AutomationError(id, 'append',
                    "{0} values given for {0} columns".format(nr_values,
                            nr_columns))

        row = table_widget.rowCount()

        QTest.mouseClick(self.add_button, Qt.LeftButton, Qt.NoModifier,
                QPoint(), delay)
        self._update_editor('append', id, delay, values[0], (nr_values == 1))

        for col_nr in range(1, nr_values):
            self._update_editor('append', id, delay, values[col_nr],
                    (col_nr == nr_values - 1))

    def simulate_move(self, id, delay, from_row, from_column, to_row, to_column):
        """ Simulate the user moving a value within the table.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param from_row:
            is the initial row of the value being moved.
        :param from_column:
            is the initial column of the value being moved.
        :param to_row:
            is the final row of the value being moved.
        :param to_column:
            is the final column of the value being moved.
        """

        # QTest doesn't support simulating DND properly (if at all).  Therefore
        # we just manipulate the underlying widget.

        itableedit = ITableEditor(self.adaptee)
        old = itableedit.value

        table_widget = self.table_widget

        self._validate_row_column('move', id, to_row, to_column)
        self._set_cell('move', id, delay, from_row, from_column)
        itm = table_widget.takeItem(from_row, from_column)
        table_widget.setItem(to_row, to_column, itm)
        self._set_cell('move', id, delay, to_row, to_column)

        new = itableedit.value
        notify_observers('value', itableedit, new, old)

    def simulate_remove(self, id, delay, row):
        """ Simulate the user removing a row from the table.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param row:
            is the row to remove.
        """

        self._set_cell('remove', id, delay, row, 0)
        QTest.mouseClick(self.remove_button, Qt.LeftButton, Qt.NoModifier,
                QPoint(), delay)

    def simulate_update(self, id, delay, value, row, column):
        """ Simulate the user updating a value in the table.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param value:
            is the new value.
        :param row:
            is the row of the value to update.
        :param column:
            is the column of the value to update.
        """

        self._set_cell('update', id, delay, row, column)
        QTest.keyClick(self.table_widget, _EDIT_KEY, Qt.NoModifier, delay)

        self._update_editor('update', id, delay, value)

    def _update_editor(self, command, id, delay, value, and_close=True):
        """ Wait for an editor to appear and use it to set the value. """

        # We can't assume either that there is an editor available, or that it
        # might soon appear.
        timeout = 50
        editor = self._editor_widget()
        while editor is None:
            if timeout < 0:
                raise AutomationError(id, command, "no editor widget found")

            QTest.qWait(5)
            timeout -= 5
            editor = self._editor_widget()

        IAutomatedEditor(editor).simulate_set(id, delay, value)
        QTest.keyClick(editor, Qt.Key_Return if and_close else Qt.Key_Tab,
                Qt.NoModifier, delay)

        if and_close:
            # Give Qt some time to tidy up and update the model.
            QTest.qWait(100)

    def _editor_widget(self):
        """ Return the widget that seems to implement the current editor. """

        # It's a shame that Qt doesn't provide an appropriate call.
        editor = QApplication.focusWidget()

        if editor is None:
            return None

        if not self.table_widget.isAncestorOf(editor):
            return None

        return editor

    def _set_cell(self, command, id, delay, row, column):
        """ Set the current cell. """

        self._validate_row_column(command, id, row, column)

        if delay >= 0:
            QTest.qWait(delay)

        # This is the easiest way to make the cell current and visible.
        self.table_widget.setCurrentCell(row, column)

    def _validate_row_column(self, command, id, row, column):
        """ Validate a row and column. """

        table_widget = self.table_widget

        if row < 0 or row >= table_widget.rowCount():
            raise AutomationError(id, command, "invalid row")

        if column < 0 or column >= table_widget.columnCount():
            raise AutomationError(id, command, "invalid column")
