# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ...ui import IView

from .object_adapter import ObjectAdapter


class LayoutAdapter(ObjectAdapter):
    """ The LayoutAdapter class is a base class for adapters that implement the
    remainder of the :class:`~dip.pui.IView` interface for Qt layouts.
    """

    @IView.enabled.getter
    def enabled(self):
        """ Invoked to get the enabled state. """

        # It is enabled if any top-level QWidget within the layout is enabled.
        for w in self.qt4_widgets(self.adaptee):
            if w.isEnabled():
                return True

        return False

    @enabled.setter
    def enabled(self, value):
        """ Invoked to set the enabled state. """

        for w in self.qt4_widgets(self.adaptee):
            w.setEnabled(value)

    @IView.status_tip.setter
    def status_tip(self, value):
        """ Invoked to set the status tip. """

        # Save for the standard getter.
        self._status_tip = value

        for w in self.qt4_widgets(self.adaptee):
            w.setStatusTip(value)

    @IView.tool_tip.setter
    def tool_tip(self, value):
        """ Invoked to set the tool tip. """

        # Save for the standard getter.
        self._tool_tip = value

        for w in self.qt4_widgets(self.adaptee):
            w.setToolTip(value)

    @IView.visible.getter
    def visible(self):
        """ Invoked to get the container's visibility. """

        # It is visible if any top-level QWidget within the layout is visible.
        for w in self.qt4_widgets(self.adaptee):
            if w.isVisible():
                return True

        return False

    @visible.setter
    def visible(self, value):
        """ Invoked to set the container's visibility. """

        for w in self.qt4_widgets(self.adaptee):
            w.setVisible(value)

    @IView.whats_this.setter
    def whats_this(self, value):
        """ Invoked to set the "What's This?" help. """

        # Save for the standard getter.
        self._whats_this = value

        for w in self.qt4_widgets(self.adaptee):
            w.setWhatsThis(value)

    def set_focus(self):
        """ Give the focus to the view. """

        views = self.views

        if len(views) > 0:
            IView(views[0]).set_focus()

    @classmethod
    def qt4_widgets(cls, layout):
        """ A generator that returns the top-level QWidgets within a layout.
        """

        for idx in range(layout.count()):
            itm = layout.itemAt(idx)

            w = itm.widget()
            if w is not None:
                yield w
            else:
                l = itm.layout()
                if l is not None:
                    for w in cls.qt4_widgets(l):
                        yield w
