# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..toolkits import Toolkit


class Application(object):
    """ The Application class creates a singleton application that implements,
    or can be adapted to the :class:`~dip.ui.IApplication` interface.
    """

    def __new__(cls, argv=None, **properties):
        """ Create the application.

        :param argv:
            is the sequence of command line arguments.  By default
            :attr:`sys.argv` is used.
        :param \*\*properties:
            are the keyword arguments used as toolkit specific property names
            and values that are used to configure the application.
        :return:
            the application.
        """

        if argv is None:
            from sys import argv

        return Toolkit.application(argv, **properties)

    @staticmethod
    def call_after(callable, after=0):
        """ Call a callable after a delay.

        :param callable:
            is the callable.
        :param after:
            is the delay in milliseconds.  If it is zero then the callable is
            called in the next iteration of the event loop.
        """

        Toolkit.call_after(callable, after)

    @staticmethod
    def information(window_title, text, parent=None):
        """ Display a informational message to the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the message.
        :param parent:
            is the optional parent view.
        """

        Toolkit.information(window_title, text, parent)

    @staticmethod
    def question(window_title, text, parent=None, detail='', buttons=('no', 'yes')):
        """ Ask the user a question.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the question.
        :param parent:
            is the optional parent view.
        :param detail:
            is the optional additional detail.
        :param buttons:
            is the sequence of buttons to display.  Possible buttons are 'yes',
            'no', 'cancel', 'save' and 'discard'.  The first in the sequence is
            used as the default.
        :return:
            The button that was pressed.
        """

        return Toolkit.question(window_title, text, parent, detail, buttons)

    @staticmethod
    def warning(window_title, text, parent=None, detail=''):
        """ Display a warning message to the user.

        :param window_title:
            is the window title, typically used as the title of a dialog.
        :param text:
            is the text of the warning.
        :param parent:
            is the optional parent view.
        :param detail:
            is the optional additional detail.
        """

        return Toolkit.warning(window_title, text, parent, detail)
