# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Any, Instance, Str
from ..toolkits import Toolkit

from .bindings import Bindings
from .i_view import IView
from .view_factory import ViewFactory


class Designer(ViewFactory):
    """ The Designer class allows any user interface created with Qt Designer
    (in the case of the default toolkit) to be used in a view.
    """

    # The bindings that connect the GUI to the model.
    bindings = Instance(Bindings)

    # The name of toolkit specific design file.
    ui_file_name = Str()

    # The toolkit specific design object.
    _ui_design = Any()

    def __init__(self, ui_file_name, **properties):
        """ Initialise the item.

        :param ui_file_name:
            is the name of the Qt Designer .ui file.
        :param \*\*properties:
            are keyword arguments used as property names and values that are
            applied to the widget.
        """

        super(Designer, self).__init__(**properties)

        # It's more natural to allow this as a positional argument.
        self.ui_file_name = ui_file_name

    def create(self, model, controller, parent):
        """ Create a view instance.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param parent:
            is the optional parent view.
        :return:
            the view that is the root of the user interface defined in the .ui
            file.
        """

        view = Toolkit.designer_view(self._ui_design, parent)

        # Bind to the model and controller.
        if self.bindings is not None:
            self.bindings.bind(view, model, controller)

        self.bind_view(view, controller)

        return view

    @_ui_design.default
    def _ui_design(self):
        """ Invoked to get the default user interface design. """

        return Toolkit.designer_load(self.ui_file_name)
