# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Model, Str, Subclass
from ..toolkits import Toolkit

from .i_object import IObject


class ObjectFactory(Model):
    """ The ObjectFactory class is the base class for all factories that
    can create an object that implements or can be adapted to the
    :class:`~dip.ui.IObject` interface.
    """

    # The object's identifier.  If it is not specified then a sub-class may
    # implement an appropriate default.
    id = Str()

    def __init__(self, **properties):
        """ Initialize the factory.

        :param \*\*properties:
            are keyword arguments used as property names and values that are
            applied to the each object created by the factory.
        """

        self._properties = properties

    def configure(self, object):
        """ Configure an object created by this factory.

        :param object:
            is the object to configure.
        """

        iobject = IObject(object)

        iobject.id = self.id
        iobject.configure(self._properties)

    @staticmethod
    def object_for_interface(interface, parent):
        """ Use the default toolkit to create an object that implements a
        particular interface.

        :param interface:
            is the interface.
        :param parent:
            is the optional parent view.
        :return:
            the object.
        """

        factory = Toolkit.object_factory_for_interface(interface)

        return factory(parent)

    @staticmethod
    def get_natural_name(name):
        """ Convert a name, typically of an attribute, to a more natural name.

        :param name:
            is the original name.
        :return:
            the more natural version.
        """

        name = name.replace('_', ' ')
        name = name.strip()
        name = name[0].upper() + name[1:]

        return name
