# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import implements, observe

from .i_container import IContainer
from .controller import Controller
from .i_editor import IEditor
from .i_wizard import IWizard
from .i_wizard_controller import IWizardController
from .i_wizard_page import IWizardPage


@implements(IWizardController)
class WizardController(Controller):
    """ The WizardController class is a :term:`controller` for handling
    wizards.
    """

    # This determines if the controller should automatically update the model
    # when the user changes the value of an editor (and the controller
    # determines that the view is valid).
    auto_update_model = False

    def validate(self):
        """ Reimplemented to disable the accept button if the view is invalid.
        """

        super(WizardController, self).validate()

        if self.current_page is not None:
            IWizardPage(self.current_page).acceptable = (self.invalid_reason == '')

    def validate_view(self):
        """ Validate the data in the view.  This implementation considers the
        view to be valid if all the editors in the current page have valid
        values.

        :return:
            a string which will be empty if the view is valid, otherwise it
            will explain why the view is invalid.
        """

        if self.current_page is None:
            return ''

        return self.validate_page(self.current_page)

    @IWizardController.view.setter
    def view(self, value):
        """ Invoked to set the view. """

        # Save for the standard getter.
        self._view = value

        iwizard = IWizard(value)

        # Observe the wizard's current page.
        observe('current_page', iwizard, self._current_page_changed)

        # Observe the wizard's finished trigger.
        observe('finished', iwizard, self._on_finished)

        # Initialise the current page.
        self.current_page = iwizard.current_page

    def validate_page(self, page):
        """ Validate a page of the view.

        :param page:
            is the page to validate.
        :return:
            a string which will be empty if the view is valid, otherwise it
            will explain why the view is invalid.
        """

        for editor in self._page_editors(page):
            ieditor = IEditor(editor)

            # Ignore disabled editors as they may legitimately contain invalid
            # values.
            if ieditor.enabled:
                invalid_reason = ieditor.invalid_reason
                if invalid_reason != '':
                    return "{0}: {1}.".format(ieditor.label, invalid_reason)

        return ''

    def _current_page_changed(self, change):
        """ Invoked when the current page changes. """

        self.current_page = change.new

        self.validate()

    def _on_finished(self, change):
        """ Invoked when the wizard's finished trigger is pulled. """

        # Update the model if not already done.
        if not self.auto_update_model:
            self.update_model()

    def _page_editors(self, page):
        """ A generator for the editors that are in a page. """

        for view in IContainer(page).views:
            if view in self.editors:
                yield view

            icontainer = IContainer(view, exception=False)
            if icontainer is not None:
                for editor in self._page_editors(view):
                    yield editor
